@@ -12,21 +12,78 @@
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
-from numba import njit
-
+from typing import List, Optional, Tuple, Union
+import numpy.typing as npt
 import numpy as np
-
-from pytensor import function as pytensor_function
-from pytensor import config
-from pytensor.tensor.var import Variable
-
-from pymc.model import modelcontext
+from numba import njit
+from pymc.model import Model, modelcontext
+from pymc.pytensorf import inputvars, join_nonshared_inputs, make_shared_replacements
 from pymc.step_methods.arraystep import ArrayStepShared
 from pymc.step_methods.compound import Competence
-from pymc.pytensorf import inputvars, join_nonshared_inputs, make_shared_replacements
+from pytensor import config
+from pytensor import function as pytensor_function
+from pytensor.tensor.var import Variable
 
 from pymc_bart.bart import BARTRV
-from pymc_bart.tree import Tree, Node, get_depth
+from pymc_bart.tree import Node, Tree, get_idx_left_child, get_idx_right_child, get_depth
+from pymc_bart.split_rules import ContinuousSplitRule
+
+
+class ParticleTree:
+    """Particle tree."""
+
+    __slots__ = "tree", "expansion_nodes", "log_weight"
+
+    def __init__(self, tree: Tree):
+        self.tree: Tree = tree.copy()
+        self.expansion_nodes: List[int] = [0]
+        self.log_weight: float = 0
+
+    def copy(self) -> "ParticleTree":
+        p = ParticleTree(self.tree)
+        p.expansion_nodes = self.expansion_nodes.copy()
+        return p
+
+    def sample_tree(
+        self,
+        ssv,
+        available_predictors,
+        prior_prob_leaf_node,
+        X,
+        missing_data,
+        sum_trees,
+        leaf_sd,
+        m,
+        response,
+        normal,
+        shape,
+    ) -> bool:
+        tree_grew = False
+        if self.expansion_nodes:
+            index_leaf_node = self.expansion_nodes.pop(0)
+            # Probability that this node will remain a leaf node
+            prob_leaf = prior_prob_leaf_node[get_depth(index_leaf_node)]
+
+            if prob_leaf < np.random.random():
+                idx_new_nodes = grow_tree(
+                    self.tree,
+                    index_leaf_node,
+                    ssv,
+                    available_predictors,
+                    X,
+                    missing_data,
+                    sum_trees,
+                    leaf_sd,
+                    m,
+                    response,
+                    normal,
+                    shape,
+                )
+                if idx_new_nodes is not None:
+                    self.expansion_nodes.extend(idx_new_nodes)
+                    tree_grew = True
+
+        return tree_grew
 
 
 class PGBART(ArrayStepShared):
@@ -38,11 +95,11 @@ class PGBART(ArrayStepShared):
     vars: list
         List of value variables for sampler
     num_particles : tuple
-        Number of particles. Defaults to 20
-    batch : int or tuple
-        Number of trees fitted per step. Defaults to  "auto", which is the 10% of the `m` trees
-        during tuning and after tuning. If a tuple is passed the first element is the batch size
-        during tuning and the second the batch size after tuning.
+        Number of particles. Defaults to 10
+    batch : tuple
+        Number of trees fitted per step. The first element is the batch size during tuning and the
+        second the batch size after tuning.  Defaults to  (0.1, 0.1), meaning 10% of the `m` trees
+        during tuning and after tuning.
     model: PyMC Model
         Optional model for sampling step. Defaults to None (taken from context).
     """
@@ -55,9 +112,9 @@ class PGBART(ArrayStepShared):
     def __init__(
         self,
         vars=None,  # pylint: disable=redefined-builtin
-        num_particles=20,
-        batch="auto",
-        model=None,
+        num_particles: int = 10,
+        batch: Tuple[float, float] = (0.1, 0.1),
+        model: Optional[Model] = None,
     ):
         model = modelcontext(model)
         initial_values = model.initial_point()
@@ -76,61 +133,82 @@ class PGBART(ArrayStepShared):
 
         self.missing_data = np.any(np.isnan(self.X))
         self.m = self.bart.m
+        self.response = self.bart.response
+
         shape = initial_values[value_bart.name].shape
-        if len(shape) == 1:
-            self.shape = 1
-        else:
-            self.shape = shape[0]
 
-        if self.bart.split_prior:
+        self.shape = 1 if len(shape) == 1 else shape[0]
+
+        # Set trees_shape (dim for separate tree structures)
+        # and leaves_shape (dim for leaf node values)
+        # One of the two is always one, the other equal to self.shape
+        self.trees_shape = self.shape if self.bart.separate_trees else 1
+        self.leaves_shape = self.shape if not self.bart.separate_trees else 1
+
+        if self.bart.split_prior.size == 0:
+            self.alpha_vec = np.ones(self.X.shape[1])
+        else:
             self.alpha_vec = self.bart.split_prior
-        else:
-            self.alpha_vec = np.ones(self.X.shape[1], dtype=np.int32)
-        init_mean = self.bart.Y.mean()
-        # if data is binary
-        y_unique = np.unique(self.bart.Y)
-        if y_unique.size == 2 and np.all(y_unique == [0, 1]):
-            mu_std = 3 / self.m**0.5
-        else:
-            mu_std = self.bart.Y.std() / self.m**0.5
 
+        if self.bart.split_rules:
+            self.split_rules = self.bart.split_rules
+        else:
+            self.split_rules = [ContinuousSplitRule] * self.X.shape[1]
+
+            jittered = np.random.normal(self.X, self.X.std(axis=0) / 12)
+            min_values = np.min(self.X, axis=0)
+            max_values = np.max(self.X, axis=0)
+            self.X = np.clip(jittered, min_values, max_values)
+
+        init_mean = self.bart.Y.mean()
         self.num_observations = self.X.shape[0]
         self.num_variates = self.X.shape[1]
         self.available_predictors = list(range(self.num_variates))
 
-        self.sum_trees = np.full((self.shape, self.bart.Y.shape[0]), init_mean).astype(
-            config.floatX
-        )
+        # if data is binary
+        self.leaf_sd = np.ones((self.trees_shape, self.leaves_shape))
+
+        y_unique = np.unique(self.bart.Y)
+        if y_unique.size == 2 and np.all(y_unique == [0, 1]):
+            self.leaf_sd *= 3 / self.m**0.5
+        else:
+            self.leaf_sd *= self.bart.Y.std() / self.m**0.5
+
+        self.running_sd = [
+            RunningSd((self.leaves_shape, self.num_observations)) for _ in range(self.trees_shape)
+        ]
+
+        self.sum_trees = np.full(
+            (self.trees_shape, self.leaves_shape, self.bart.Y.shape[0]), init_mean
+        ).astype(config.floatX)
         self.sum_trees_noi = self.sum_trees - init_mean
         self.a_tree = Tree.new_tree(
             leaf_node_value=init_mean / self.m,
             idx_data_points=np.arange(self.num_observations, dtype="int32"),
             num_observations=self.num_observations,
-            shape=self.shape,
+            shape=self.leaves_shape,
+            split_rules=self.split_rules,
         )
-        self.normal = NormalSampler(mu_std, self.shape)
+
+        self.normal = NormalSampler(1, self.leaves_shape)
         self.uniform = UniformSampler(0, 1)
-        self.uniform_kf = UniformSampler(0.33, 0.75, self.shape)
-        self.prior_prob_leaf_node = compute_prior_probability(self.bart.alpha)
+        self.prior_prob_leaf_node = compute_prior_probability(self.bart.alpha, self.bart.beta)
         self.ssv = SampleSplittingVariable(self.alpha_vec)
 
         self.tune = True
 
-        if batch == "auto":
-            batch = max(1, int(self.m * 0.1))
-            self.batch = (batch, batch)
-        else:
-            if isinstance(batch, (tuple, list)):
-                self.batch = batch
-            else:
-                self.batch = (batch, batch)
+        batch_0 = max(1, int(self.m * batch[0]))
+        batch_1 = max(1, int(self.m * batch[1]))
+        self.batch = (batch_0, batch_1)
 
         self.num_particles = num_particles
         self.indices = list(range(1, num_particles))
         shared = make_shared_replacements(initial_values, vars, model)
         self.likelihood_logp = logp(initial_values, [model.datalogp], vars, shared)
-        self.all_particles = list(ParticleTree(self.a_tree) for _ in range(self.m))
-        self.all_trees = np.array([p.tree for p in self.all_particles])
+        self.all_particles = [
+            [ParticleTree(self.a_tree) for _ in range(self.m)] for _ in range(self.trees_shape)
+        ]
+        self.all_trees = np.array([[p.tree for p in pl] for pl in self.all_particles])
         self.lower = 0
         self.iter = 0
         super().__init__(vars, shared)
@@ -142,62 +220,75 @@ class PGBART(ArrayStepShared):
         tree_ids = range(self.lower, upper)
         self.lower = upper if upper < self.m else 0
 
-        for tree_id in tree_ids:
-            self.iter += 1
-            # Compute the sum of trees without the old tree that we are attempting to replace
-            self.sum_trees_noi = self.sum_trees - self.all_particles[tree_id].tree._predict()
-            # Generate an initial set of particles
-            # at the end we return one of these particles as the new tree
-            particles = self.init_particles(tree_id)
+        for odim in range(self.trees_shape):
+            for tree_id in tree_ids:
+                self.iter += 1
+                # Compute the sum of trees without the old tree that we are attempting to replace
+                self.sum_trees_noi[odim] = (
+                    self.sum_trees[odim] - self.all_particles[odim][tree_id].tree._predict()
+                )
+                # Generate an initial set of particles
+                # at the end we return one of these particles as the new tree
+                particles = self.init_particles(tree_id, odim)
 
-            while True:
-                # Sample each particle (try to grow each tree), except for the first one
-                stop_growing = True
-                for p in particles[1:]:
-                    tree_grew = p.sample_tree(
-                        self.ssv,
-                        self.available_predictors,
-                        self.prior_prob_leaf_node,
-                        self.X,
-                        self.missing_data,
-                        self.sum_trees,
-                        self.m,
-                        self.normal,
-                        self.shape,
-                    )
-                    if tree_grew:
-                        self.update_weight(p)
-                    if p.expansion_nodes:
-                        stop_growing = False
-                if stop_growing:
-                    break
+                while True:
+                    # Sample each particle (try to grow each tree), except for the first one
+                    stop_growing = True
+                    for p in particles[1:]:
+                        if p.sample_tree(
+                            self.ssv,
+                            self.available_predictors,
+                            self.prior_prob_leaf_node,
+                            self.X,
+                            self.missing_data,
+                            self.sum_trees[odim],
+                            self.leaf_sd[odim],
+                            self.m,
+                            self.response,
+                            self.normal,
+                            self.leaves_shape,
+                        ):
+                            self.update_weight(p, odim)
+                        if p.expansion_nodes:
+                            stop_growing = False
+                    if stop_growing:
+                        break
 
-                # Normalize weights
-                normalized_weights = self.normalize(particles[1:])
+                    # Normalize weights
+                    normalized_weights = self.normalize(particles[1:])
 
-                # Resample
-                particles = self.resample(particles, normalized_weights)
+                    # Resample
+                    particles = self.resample(particles, normalized_weights)
 
-            normalized_weights = self.normalize(particles)
-            # Get the new particle and associated tree
-            self.all_particles[tree_id], new_tree = self.get_particle_tree(
-                particles, normalized_weights
-            )
-            # Update the sum of trees
-            self.sum_trees = self.sum_trees_noi + new_tree._predict()
-            # To reduce memory usage, we trim the tree
-            self.all_trees[tree_id] = new_tree.trim()
+                normalized_weights = self.normalize(particles)
+                # Get the new particle and associated tree
+                self.all_particles[odim][tree_id], new_tree = self.get_particle_tree(
+                    particles, normalized_weights
+                )
+                # Update the sum of trees
+                new = new_tree._predict()
+                self.sum_trees[odim] = self.sum_trees_noi[odim] + new
+                # To reduce memory usage, we trim the tree
+                self.all_trees[odim][tree_id] = new_tree.trim()
 
-            if self.tune:
-                # Update the splitting variable and the splitting variable sampler
-                if self.iter > self.m:
-                    self.ssv = SampleSplittingVariable(self.alpha_vec)
-                for index in new_tree.get_split_variables():
-                    self.alpha_vec[index] += 1
-            else:
-                # update the variable inclusion
-                for index in new_tree.get_split_variables():
-                    variable_inclusion[index] += 1
+                if self.tune:
+                    # Update the splitting variable and the splitting variable sampler
+                    if self.iter > self.m:
+                        self.ssv = SampleSplittingVariable(self.alpha_vec)
+
+                    for index in new_tree.get_split_variables():
+                        self.alpha_vec[index] += 1
+
+                    # update standard deviation at leaf nodes
+                    if self.iter > 2:
+                        self.leaf_sd[odim] = self.running_sd[odim].update(new)
+                    else:
+                        self.running_sd[odim].update(new)
+
+                else:
+                    # update the variable inclusion
+                    for index in new_tree.get_split_variables():
+                        variable_inclusion[index] += 1
 
         if not self.tune:
             self.bart.all_trees.append(self.all_trees)
@@ -205,7 +296,7 @@ class PGBART(ArrayStepShared):
         stats = {"variable_inclusion": variable_inclusion, "tune": self.tune}
         return self.sum_trees, [stats]
 
-    def normalize(self, particles):
+    def normalize(self, particles: List[ParticleTree]) -> float:
         """
         Use softmax to get normalized_weights.
         """
@@ -215,15 +306,17 @@ class PGBART(ArrayStepShared):
         wei = np.exp(log_w_) + 1e-12
         return wei / wei.sum()
 
-    def resample(self, particles, normalized_weights):
+    def resample(
+        self, particles: List[ParticleTree], normalized_weights: npt.NDArray[np.float_]
+    ) -> List[ParticleTree]:
         """
         Use systematic resample for all but the first particle
 
         Ensure particles are copied only if needed.
         """
         new_indices = self.systematic(normalized_weights) + 1
-        seen = []
-        new_particles = []
+        seen: List[int] = []
+        new_particles: List[ParticleTree] = []
         for idx in new_indices:
             if idx in seen:
                 new_particles.append(particles[idx].copy())
@@ -235,7 +328,9 @@ class PGBART(ArrayStepShared):
 
         return particles
 
-    def get_particle_tree(self, particles, normalized_weights):
+    def get_particle_tree(
+        self, particles: List[ParticleTree], normalized_weights: npt.NDArray[np.float_]
+    ) -> Tuple[ParticleTree, Tree]:
         """
         Sample a new particle and associated tree
         """
@@ -246,7 +341,7 @@ class PGBART(ArrayStepShared):
 
         return new_particle, new_particle.tree
 
-    def systematic(self, normalized_weights):
+    def systematic(self, normalized_weights: npt.NDArray[np.float_]) -> npt.NDArray[np.int_]:
         """
         Systematic resampling.
 
@@ -258,27 +353,27 @@ class PGBART(ArrayStepShared):
         single_uniform = (self.uniform.rvs() + np.arange(lnw)) / lnw
         return inverse_cdf(single_uniform, normalized_weights)
 
-    def init_particles(self, tree_id: int) -> np.ndarray:
+    def init_particles(self, tree_id: int, odim: int) -> List[ParticleTree]:
         """Initialize particles."""
-        p0 = self.all_particles[tree_id]
+        p0: ParticleTree = self.all_particles[odim][tree_id]
         # The old tree does not grow so we update the weight only once
-        self.update_weight(p0)
-        particles = [p0]
-
-        for _ in self.indices:
-            particles.append(
-                ParticleTree(self.a_tree, self.uniform_kf.rvs() if self.tune else p0.kfactor)
-            )
+        self.update_weight(p0, odim)
+        particles: List[ParticleTree] = [p0]
 
+        particles.extend(ParticleTree(self.a_tree) for _ in self.indices)
         return particles
 
-    def update_weight(self, particle):
+    def update_weight(self, particle: ParticleTree, odim: int) -> None:
         """
         Update the weight of a particle.
         """
-        new_likelihood = self.likelihood_logp(
-            (self.sum_trees_noi + particle.tree._predict()).flatten()
+
+        delta = (
+            np.identity(self.trees_shape)[odim][:, None, None]
+            * particle.tree._predict()[None, :, :]
         )
+
+        new_likelihood = self.likelihood_logp((self.sum_trees_noi + delta).flatten())
         particle.log_weight = new_likelihood
 
     @staticmethod
@@ -290,64 +385,38 @@ class PGBART(ArrayStepShared):
         return Competence.INCOMPATIBLE
 
 
-class ParticleTree:
-    """Particle tree."""
+class RunningSd:
+    """Welford's online algorithm for computing the variance/standard deviation"""
 
-    __slots__ = "tree", "expansion_nodes", "log_weight", "kfactor"
+    def __init__(self, shape: tuple) -> None:
+        self.count = 0  # number of data points
+        self.mean = np.zeros(shape)  # running mean
+        self.m_2 = np.zeros(shape)  # running second moment
 
-    def __init__(self, tree, kfactor=0.75):
-        self.tree = tree.copy()
-        self.expansion_nodes = [0]
-        self.log_weight = 0
-        self.kfactor = kfactor
+    def update(self, new_value: npt.NDArray[np.float_]) -> Union[float, npt.NDArray[np.float_]]:
+        self.count = self.count + 1
+        self.mean, self.m_2, std = _update(self.count, self.mean, self.m_2, new_value)
+        return fast_mean(std)
 
-    def copy(self):
-        p = ParticleTree(self.tree)
-        p.expansion_nodes = self.expansion_nodes.copy()
-        p.kfactor = self.kfactor
-        return p
 
-    def sample_tree(
-        self,
-        ssv,
-        available_predictors,
-        prior_prob_leaf_node,
-        X,
-        missing_data,
-        sum_trees,
-        m,
-        normal,
-        shape,
-    ):
-        tree_grew = False
-        if self.expansion_nodes:
-            index_leaf_node = self.expansion_nodes.pop(0)
-            # Probability that this node will remain a leaf node
-            prob_leaf = prior_prob_leaf_node[get_depth(index_leaf_node)]
+@njit
+def _update(
+    count: int,
+    mean: npt.NDArray[np.float_],
+    m_2: npt.NDArray[np.float_],
+    new_value: npt.NDArray[np.float_],
+) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], Union[float, npt.NDArray[np.float_]]]:
+    delta = new_value - mean
+    mean += delta / count
+    delta2 = new_value - mean
+    m_2 += delta * delta2
 
-            if prob_leaf < np.random.random():
-                idx_new_nodes = grow_tree(
-                    self.tree,
-                    index_leaf_node,
-                    ssv,
-                    available_predictors,
-                    X,
-                    missing_data,
-                    sum_trees,
-                    m,
-                    normal,
-                    self.kfactor,
-                    shape,
-                )
-                if idx_new_nodes is not None:
-                    self.expansion_nodes.extend(idx_new_nodes)
-                    tree_grew = True
-
-        return tree_grew
+    std = (m_2 / count) ** 0.5
+    return mean, m_2, std
 
 
 class SampleSplittingVariable:
-    def __init__(self, alpha_vec):
+    def __init__(self, alpha_vec: npt.NDArray[np.float_]) -> None:
         """
         Sample splitting variables proportional to `alpha_vec`.
 
@@ -356,38 +425,34 @@ class SampleSplittingVariable:
         """
         self.enu = list(enumerate(np.cumsum(alpha_vec / alpha_vec.sum())))
 
-    def rvs(self):
-        rnd = np.random.random()
+    def rvs(self) -> Union[int, Tuple[int, float]]:
+        rnd: float = np.random.random()
         for i, val in self.enu:
             if rnd <= val:
                 return i
         return self.enu[-1]
 
 
-def compute_prior_probability(alpha):
+def compute_prior_probability(alpha: int, beta: int) -> List[float]:
     """
     Calculate the probability of the node being a leaf node (1 - p(being split node)).
 
-    Taken from equation 19 in [Rockova2018].
-
     Parameters
     ----------
     alpha : float
+    beta: float
 
     Returns
     -------
     list with probabilities for leaf nodes
-
-    References
-    ----------
-    .. [Rockova2018] Veronika Rockova, Enakshi Saha (2018). On the theory of BART.
-    arXiv, `link <https://arxiv.org/abs/1810.00787>`__
     """
-    prior_leaf_prob = [0]
-    depth = 1
-    while prior_leaf_prob[-1] < 1:
-        prior_leaf_prob.append(1 - alpha**depth)
+    prior_leaf_prob: List[float] = [0]
+    depth = 0
+    while prior_leaf_prob[-1] < 0.9999:
+        prior_leaf_prob.append(1 - (alpha * ((1 + depth) ** (-beta))))
         depth += 1
+    prior_leaf_prob.append(1)
+
     return prior_leaf_prob
 
 
@@ -399,9 +464,10 @@ def grow_tree(
     X,
     missing_data,
     sum_trees,
+    leaf_sd,
     m,
+    response,
     normal,
-    kfactor,
     shape,
 ):
     current_node = tree.get_node(index_leaf_node)
@@ -409,82 +475,87 @@ def grow_tree(
 
     index_selected_predictor = ssv.rvs()
     selected_predictor = available_predictors[index_selected_predictor]
-    available_splitting_values = X[idx_data_points, selected_predictor]
-    split_value = get_split_value(available_splitting_values, idx_data_points, missing_data)
+    idx_data_points, available_splitting_values = filter_missing_values(
+        X[idx_data_points, selected_predictor], idx_data_points, missing_data
+    )
+
+    split_rule = tree.split_rules[selected_predictor]
+
+    split_value = split_rule.get_split_value(available_splitting_values)
 
     if split_value is None:
         return None
-    new_idx_data_points = get_new_idx_data_points(
-        available_splitting_values, split_value, idx_data_points
-    )
+
+    to_left = split_rule.divide(available_splitting_values, split_value)
+    new_idx_data_points = idx_data_points[to_left], idx_data_points[~to_left]
+
     current_node_children = (
-        current_node.get_idx_left_child(),
-        current_node.get_idx_right_child(),
+        get_idx_left_child(index_leaf_node),
+        get_idx_right_child(index_leaf_node),
     )
 
-    new_nodes = []
+    if response == "mix":
+        response = "linear" if np.random.random() >= 0.5 else "constant"
+
     for idx in range(2):
         idx_data_point = new_idx_data_points[idx]
-        node_value = draw_leaf_value(
-            sum_trees[:, idx_data_point],
-            m,
-            normal.rvs() * kfactor,
-            shape,
+        node_value, linear_params = draw_leaf_value(
+            y_mu_pred=sum_trees[:, idx_data_point],
+            x_mu=X[idx_data_point, selected_predictor],
+            m=m,
+            norm=normal.rvs() * leaf_sd,
+            shape=shape,
+            response=response,
         )
 
         new_node = Node.new_leaf_node(
-            index=current_node_children[idx],
             value=node_value,
+            nvalue=len(idx_data_point),
             idx_data_points=idx_data_point,
+            linear_params=linear_params,
         )
-        new_nodes.append(new_node)
+        tree.set_node(current_node_children[idx], new_node)
 
     tree.grow_leaf_node(current_node, selected_predictor, split_value, index_leaf_node)
-    tree.set_node(new_nodes[0].index, new_nodes[0])
-    tree.set_node(new_nodes[1].index, new_nodes[1])
-
-    return [new_nodes[0].index, new_nodes[1].index]
+    return current_node_children
 
 
-@njit
-def get_new_idx_data_points(available_splitting_values, split_value, idx_data_points):
-    split_idx = available_splitting_values <= split_value
-    return idx_data_points[split_idx], idx_data_points[~split_idx]
-
-
-def get_split_value(available_splitting_values, idx_data_points, missing_data):
+def filter_missing_values(available_splitting_values, idx_data_points, missing_data):
     if missing_data:
-        idx_data_points = idx_data_points[~np.isnan(available_splitting_values)]
-        available_splitting_values = available_splitting_values[
-            ~np.isnan(available_splitting_values)
-        ]
-
-    split_value = None
-    if available_splitting_values.size > 0:
-        idx_selected_splitting_values = discrete_uniform_sampler(len(available_splitting_values))
-        split_value = available_splitting_values[idx_selected_splitting_values]
-
-    return split_value
+        mask = ~np.isnan(available_splitting_values)
+        idx_data_points = idx_data_points[mask]
+        available_splitting_values = available_splitting_values[mask]
+    return idx_data_points, available_splitting_values
 
 
-@njit
-def draw_leaf_value(y_mu_pred, m, norm, shape):
+def draw_leaf_value(
+    y_mu_pred: npt.NDArray[np.float_],
+    x_mu: npt.NDArray[np.float_],
+    m: int,
+    norm: npt.NDArray[np.float_],
+    shape: int,
+    response: str,
+) -> Tuple[npt.NDArray[np.float_], Optional[npt.NDArray[np.float_]]]:
     """Draw Gaussian distributed leaf values."""
+    linear_params = None
+    mu_mean = np.empty(shape)
     if y_mu_pred.size == 0:
-        return np.zeros(shape)
+        return np.zeros(shape), linear_params
 
     if y_mu_pred.size == 1:
-        mu_mean = np.full(shape, y_mu_pred.item() / m)
+        mu_mean = np.full(shape, y_mu_pred.item() / m) + norm
     else:
-        mu_mean = fast_mean(y_mu_pred) / m
+        if y_mu_pred.size < 3 or response == "constant":
+            mu_mean = fast_mean(y_mu_pred) / m + norm
+        else:
+            mu_mean, linear_params = fast_linear_fit(x=x_mu, y=y_mu_pred, m=m, norm=norm)
 
-    return norm + mu_mean
+    return mu_mean, linear_params
 
 
 @njit
-def fast_mean(ari):
+def fast_mean(ari: npt.NDArray[np.float_]) -> Union[float, npt.NDArray[np.float_]]:
     """Use Numba to speed up the computation of the mean."""
-
     if ari.ndim == 1:
         count = ari.shape[0]
         suma = 0
@@ -500,6 +571,35 @@ def fast_mean(ari):
         return res / count
 
 
+@njit
+def fast_linear_fit(
+    x: npt.NDArray[np.float_],
+    y: npt.NDArray[np.float_],
+    m: int,
+    norm: npt.NDArray[np.float_],
+) -> Tuple[npt.NDArray[np.float_], List[npt.NDArray[np.float_]]]:
+    n = len(x)
+    y = y / m + np.expand_dims(norm, axis=1)
+
+    xbar = np.sum(x) / n
+    ybar = np.sum(y, axis=1) / n
+
+    x_diff = x - xbar
+    y_diff = y - np.expand_dims(ybar, axis=1)
+
+    x_var = np.dot(x_diff, x_diff.T)
+
+    if x_var == 0:
+        b = np.zeros(y.shape[0])
+    else:
+        b = np.dot(x_diff, y_diff.T) / x_var
+
+    a = ybar - b * xbar
+
+    y_fit = np.expand_dims(a, axis=1) + np.expand_dims(b, axis=1) * x
+    return y_fit.T, [a, b]
+
+
 def discrete_uniform_sampler(upper_value):
     """Draw from the uniform distribution with bounds [0, upper_value).
 
@@ -560,17 +660,19 @@ class UniformSampler:
 
 
 @njit
-def inverse_cdf(single_uniform, normalized_weights):
+def inverse_cdf(
+    single_uniform: npt.NDArray[np.float_], normalized_weights: npt.NDArray[np.float_]
+) -> npt.NDArray[np.int_]:
     """
     Inverse CDF algorithm for a finite distribution.
 
     Parameters
     ----------
-    single_uniform: ndarray
-        ordered points in [0,1]
+    single_uniform: npt.NDArray[np.float_]
+        Ordered points in [0,1]
 
-    normalized_weights: ndarray
-        normalized weights
+    normalized_weights: npt.NDArray[np.float_])
+        Normalized weights
 
     Returns
     -------
